/*
 * Copyright (c) 2005 MontaVista Software, Inc.
 * Copyright (c) 2006-2012 Red Hat, Inc.
 *
 * All rights reserved.
 *
 * Author: Steven Dake (sdake@redhat.com)

 * This software licensed under BSD license, the text of which follows:
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the MontaVista Software, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config.h>

#include <assert.h>
#include <pthread.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/un.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sched.h>
#include <time.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/uio.h>
#include <limits.h>
#include <netdb.h>
#include <sys/vfs.h>
#include <linux/fs.h>
#include <corosync/sq.h>
#include <corosync/swab.h>
#include <corosync/list.h>
#include <qb/qbdefs.h>
#include <qb/qbloop.h>
#define LOGSYS_UTILS_ONLY 1
#include <corosync/logsys.h>
#include "totemdisk.h"

#include "util.h"
#include "totemcrypto.h"

#include <nss.h>
#include <pk11pub.h>
#include <pkcs11.h>
#include <prerror.h>
#include <corosync/icmap.h>

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif

#define MCAST_SOCKET_BUFFER_SIZE (TRANSMITS_ALLOWED * FRAME_SIZE_MAX)
#define NETIF_STATE_REPORT_UP		1
#define NETIF_STATE_REPORT_DOWN		2

#define BIND_STATE_UNBOUND	0
#define BIND_STATE_REGULAR	1
#define BIND_STATE_LOOPBACK	2

//#define MESSAGE_TYPE_MEMB_JOIN	3
#define SECTOR_SIZE   100 
#define MAX_NODE_NUM  16
#define DISK_LEN  65564
#define MSG_LEN 65536

#define ENDIAN_LOCAL                                     0xff22

enum message_type {
        MESSAGE_TYPE_ORF_TOKEN = 0,                     /* Ordering, Reliability, Flow (ORF) control Token */
        MESSAGE_TYPE_MCAST = 1,                         /* ring ordered multicast message */
        MESSAGE_TYPE_MEMB_MERGE_DETECT = 2,     /* merge rings if there are available rings */
        MESSAGE_TYPE_MEMB_JOIN = 3,                     /* membership join message */
        MESSAGE_TYPE_MEMB_COMMIT_TOKEN = 4,     /* membership commit token */
        MESSAGE_TYPE_TOKEN_HOLD_CANCEL = 5,     /* cancel the holding of the token */
};

enum encapsulation_type {
        MESSAGE_ENCAPSULATED = 1,
        MESSAGE_NOT_ENCAPSULATED = 2
};


struct totemdisk_socket {
	/*
	 * Socket used for local multicast delivery. We don't rely on multicast
	 * loop and rather this UNIX DGRAM socket is used. Socket is created by
	 * socketpair call and they are used in same way as pipe (so [0] is read
	 * end and [1] is write end)
	 */
//	int disk_fd;
	int local_mcast_loop[2];
};

typedef struct{
	char name[4];
	int age;
}people;
int first = 0;
char temp = 'a';

struct totemdisk_instance {
	struct crypto_instance *crypto_inst;

	qb_loop_t *totemdisk_poll_handle;

	struct totem_interface *totem_interface;

	int netif_state_report;

	int netif_bind_state;

	void *context;

	void (*totemdisk_deliver_fn) (
		void *context,
		const void *msg,
		unsigned int msg_len);

	void (*totemdisk_iface_change_fn) (
		void *context,
		const struct totem_ip_address *iface_address);

	void (*totemdisk_target_set_completed) (void *context);

	/*
	 * Function and data used to log messages
	 */
	int totemdisk_log_level_security;

	int totemdisk_log_level_error;

	int totemdisk_log_level_warning;

	int totemdisk_log_level_notice;

	int totemdisk_log_level_debug;

	int totemdisk_subsys_id;

	void (*totemdisk_log_printf) (
		int level,
		int subsys,
		const char *function,
		const char *file,
		int line,
		const char *format,
		...)__attribute__((format(printf, 6, 7)));

	void *disk_context;

	char iov_buffer[FRAME_SIZE_MAX];

	char iov_buffer_flush[FRAME_SIZE_MAX];

	struct iovec totemdisk_iov_recv;

	struct iovec totemdisk_iov_recv_flush;

	struct totemdisk_socket totemdisk_sockets;

	struct totem_ip_address mcast_address;

	int stats_sent;

	int stats_recv;

	int stats_delv;

	int stats_remcasts;

	int stats_orf_token;

	struct timeval stats_tv_start;

	struct totem_ip_address my_id;

	int firstrun;

	qb_loop_timer_handle timer_netif_check_timeout;

	qb_loop_timer_handle timer_netif_disk_read;

	unsigned int my_memb_entries;

	int flushing;

	struct totem_config *totem_config;

	totemsrp_stats_t *stats;

	struct totem_ip_address token_target;

	unsigned int target_id;
};

struct work_item {
	const void *msg;
	unsigned int msg_len;
	struct totemdisk_instance *instance;
};

static void diskif_determine (
        struct totemdisk_instance *instance,
        int *interface_up);

static struct totem_ip_address localhost;

//typedef 
struct heart_info
{
	char isuse[10];//该扇区是否被节点使用
	unsigned int node_id;//id
	time_t timestamp;//当前时间戳
        char msg[MSG_LEN];
} __attribute__((packed));

int count = 0; //一个节点写一块区域 存十个消息
long int pre_time[100]= {0};//记录前一次的时间戳
long int pre_sec[100] ={0};
long int pre_usec[100] = {0}; 
int check_flag = 0;//检查一次即可
unsigned int sector_node = 0; //本节点写磁盘的区域;
unsigned int read_error_flag = 0;
int read_sector[SECTOR_SIZE] = {0};

struct message_header {
	char type;
	char encapsulated;
	unsigned short endian_detector;
	unsigned int nodeid;
} __attribute__((packed));

struct disk_msg{
	struct message_header header;
	unsigned int node_id;
	unsigned int timestamp;
}_diskmsg_;

#define log_printf(level, format, args...)                              \
do {                                                                    \
        instance->totemdisk_log_printf (                                        \
                level, instance->totemdisk_subsys_id,                   \
                __FUNCTION__, __FILE__, __LINE__,                       \
                (const char *)format, ##args);                          \
} while (0);

#define LOGSYS_PERROR(err_num, level, fmt, args...)                                             \
do {                                                                                            \
        char _error_str[LOGSYS_MAX_PERROR_MSG_LEN];                                             \
        const char *_error_ptr = qb_strerror_r(err_num, _error_str, sizeof(_error_str));        \
        instance->totemdisk_log_printf (                                                                \
                level, instance->totemdisk_subsys_id,                                           \
                __FUNCTION__, __FILE__, __LINE__,                                               \
                fmt ": %s (%d)\n", ##args, _error_ptr, err_num);                                \
        } while(0)

static ssize_t socket_write(struct totemdisk_instance *instance, const void *buf, size_t len) 
{
        char sendbuf[65536 + sizeof(unsigned int)];

        *(unsigned int *)sendbuf = len;
        memcpy(sendbuf + sizeof(unsigned int), buf, len);
        return write(instance->totemdisk_sockets.local_mcast_loop[1], sendbuf, len + sizeof(unsigned int));
}

static ssize_t disk_recvmsg(int fd, void *buf, size_t len)
{
        unsigned int msg_len;
	int res;

        res = read(fd, &msg_len, sizeof(unsigned int));

        if (msg_len > 65536) {
                return -1;
        }

        if (msg_len == 0) {
                return 0;
        }

        return read(fd, buf, msg_len);
}

static inline void write_disk(struct totemdisk_instance *instance, const void *msg, unsigned int msg_len)
{
	int fd,res;
	time_t t;
	char buf[DISK_LEN];
	struct timeval tv;
	time(&t);
	gettimeofday(&tv,NULL);
	if(msg_len > MSG_LEN)
		return;

	memset(buf, 0, sizeof(buf));
	*(unsigned int * )buf = 1;
	*((unsigned int *)(buf + 4)) = instance->totem_config->node_id;
	memcpy(buf + 2 * sizeof(int), &tv, sizeof(struct timeval));
	memcpy(buf + 2 * sizeof(int) + sizeof(struct timeval), &msg_len, sizeof(int));
	memcpy(buf + 3 * sizeof(int) + sizeof(struct timeval), msg, msg_len);

	fd = open(instance->totem_config->diskdev, O_WRONLY|O_SYNC);
	if(fd < 0)
		LOGSYS_PERROR (errno, instance->totemdisk_log_level_debug, "aaaaa open fail");

	res = pwrite(fd, buf, DISK_LEN, (sector_node*SECTOR_SIZE + count) * DISK_LEN);
	if(res < 0)
		LOGSYS_PERROR (errno, instance->totemdisk_log_level_debug, "aaaaa write fail");
	else {
		if(count == (SECTOR_SIZE - 1))
			count = 0;
		else
			count++;
	}
	fdatasync(fd);
	close(fd);
}

static inline void cast_sendmsg (
        struct totemdisk_instance *instance,
//        struct totem_ip_address *system_to,
        const void *msg,
        unsigned int msg_len)
{
	int res;//, fd;
	size_t buf_out_len;
	unsigned char buf_out[FRAME_SIZE_MAX];
	struct iovec iovec;

	/*
         * Encrypt and digest the message
         */
        if (crypto_encrypt_and_sign (
                instance->crypto_inst,
		(const unsigned char *)(msg),
		msg_len,
                buf_out,
                &buf_out_len) != 0) {
                log_printf(LOGSYS_LEVEL_CRIT, "Error encrypting/signing packet (non-critical)");
                return;
        }

	iovec.iov_base = (void *)buf_out;
	iovec.iov_len = buf_out_len;

        res = socket_write(instance, iovec.iov_base, iovec.iov_len);
	if (res < 0) {
		LOGSYS_PERROR (errno, instance->totemdisk_log_level_debug,
			"sendmsg(local mcast loop) failed (non-critical)");
	}
}

static void totemdisk_instance_initialize (struct totemdisk_instance *instance)
{
	memset (instance, 0, sizeof (struct totemdisk_instance));

	instance->netif_state_report = NETIF_STATE_REPORT_UP | NETIF_STATE_REPORT_DOWN;

	instance->totemdisk_iov_recv.iov_base = instance->iov_buffer;

	instance->totemdisk_iov_recv.iov_len = FRAME_SIZE_MAX; //sizeof (instance->iov_buffer);
	instance->totemdisk_iov_recv_flush.iov_base = instance->iov_buffer_flush;

	instance->totemdisk_iov_recv_flush.iov_len = FRAME_SIZE_MAX; //sizeof (instance->iov_buffer);

	/*
	 * There is always atleast 1 processor
	 */
	instance->my_memb_entries = 1;
}

int totemdisk_crypto_set (
	void *disk_context,
	const char *cipher_type,
	const char *hash_type)
{

	return (0);
}

int totemdisk_finalize (
	void *disk_context)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	int res = 0;

        if (instance->totemdisk_sockets.local_mcast_loop[0] > 0) {
                qb_loop_poll_del(instance->totemdisk_poll_handle,
                        instance->totemdisk_sockets.local_mcast_loop[0]);
                close (instance->totemdisk_sockets.local_mcast_loop[0]);
        }

        if (instance->totemdisk_sockets.local_mcast_loop[1] > 0) {
                close (instance->totemdisk_sockets.local_mcast_loop[1]);
	}

	return (res);
}

/*
 * Only designed to work with a message with one iov
 */

static int disk_deliver_fn (
	int fd,
	int revents,
	void *data)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)data;
	//struct msghdr msg_recv;
	struct iovec *iovec;
	//struct sockaddr_storage system_from;
	int bytes_received;
	char *message_type;
	int res = 0;

	if (instance->flushing == 1) {
		iovec = &instance->totemdisk_iov_recv_flush;
	} else {
		iovec = &instance->totemdisk_iov_recv;
	}

	bytes_received = disk_recvmsg (instance->totemdisk_sockets.local_mcast_loop[0], iovec->iov_base, iovec->iov_len);

	if (bytes_received <= 0) {
		return (0);
	} else {
		instance->stats_recv += bytes_received;
	}

       /*
        * Authenticate and if authenticated, decrypt datagram
        */
	res = crypto_authenticate_and_decrypt (instance->crypto_inst, iovec->iov_base, &bytes_received);
	if (res == -1) {
               log_printf (instance->totemdisk_log_level_security, "Received message has invalid digest... ignoring.");
               log_printf (instance->totemdisk_log_level_security,
                       "Invalid packet data");
               iovec->iov_len = FRAME_SIZE_MAX;
               return 0;
	}
	iovec->iov_len = bytes_received;


	message_type = (char *)iovec->iov_base;
	if (instance->flushing == 1 && *message_type == MESSAGE_TYPE_MEMB_JOIN) {
		iovec->iov_len = FRAME_SIZE_MAX;
		return (0);
	}

	/*
	 * Handle incoming message
	 */
	instance->totemdisk_deliver_fn (
		instance->context,
		iovec->iov_base,
		iovec->iov_len);

	iovec->iov_len = FRAME_SIZE_MAX;
        return (0);

}

static void diskif_determine (
	struct totemdisk_instance *instance,
	int *interface_up)
{
	int res;

        res = totemdisk_check(instance->totem_config->diskdev);
      	if(res == -1)
	{
		*interface_up = 0;
	}else{
		*interface_up = 1;
	}
	return;
}

static int disk_init(struct totemdisk_instance *instance)
{
	int dev;
	char buf[DISK_LEN];
	//char buf_init[DISK_LEN * SECTOR_SIZE];
	int ret = 0;
	int i = 0; 
	int result =  -1 ;
	unsigned int use, id, my_id;

	my_id = instance->totem_config->node_id;
	//memset(buf_init,0,sizeof(buf_init));

	dev = open(instance->totem_config->diskdev, O_RDWR);
	assert(dev != -1);
	for(i = 0;i < MAX_NODE_NUM;i++)
	{
		if(pread(dev,buf,sizeof(buf),DISK_LEN*SECTOR_SIZE*i) == -1)
		{
			LOGSYS_PERROR (errno, instance->totemdisk_log_level_debug, "read error");
			close(dev);
			return -1;
		}
		use = *(unsigned int *)buf;
		id = *(unsigned int *)(buf+4);
		if(use == 1)
		{
			if(id == my_id)
			{
				sector_node = i;
				result = i;
				close(dev);
				return result;
			}
		}else{
			sector_node = i;
			result = i;

			ret = pwrite(dev, 0x0, DISK_LEN*SECTOR_SIZE, DISK_LEN * SECTOR_SIZE * i);
			if(ret== -1){
				LOGSYS_PERROR (errno, instance->totemdisk_log_level_debug, "write error");
				close(dev);
				return -1;
			}
			close(dev); 
			return result;
		}

	}
	close(dev);
	return -1;
}

static void timer_function_disk_read_timeout (void *data) {

	struct totemdisk_instance *instance = data;
	int fd;
	int interface_up = 1;
	int res = 0;
	char buf[DISK_LEN];
	int i = 0, j = 0;
        struct iovec iovec;
        size_t buf_out_len;
        unsigned char buf_out[FRAME_SIZE_MAX];
//	unsigned int id = 0 ;
        time_t timestamp = 0;
	struct timeval tv;
        unsigned int msg_len;
	int read_max_sector = 0;
	const char * readdiskdev[2]={"/dev/sdb", "/dev/sdc"};

	//for (j = 0; j< 2; j++){
		fd = open(instance->totem_config->diskdev, O_RDONLY);
		//fd = open(readdiskdev[j], O_RDONLY);
		if (fd < 0) {
			diskif_determine(instance, &interface_up);
			LOGSYS_PERROR (errno, instance->totemdisk_log_level_debug, "open disk error");
		//	continue;
		//}
	}else{
		for(j = 0; j < 3; j++){
			if(j == sector_node)
				continue;
			read_error_flag = 0;

			read_max_sector = 0;
			for(i= 0; i < 100; i++)
			{
    				memset(buf, 0, DISK_LEN);
				if(pread(fd, buf, DISK_LEN, (j*SECTOR_SIZE+(i + read_sector[j])%SECTOR_SIZE)*DISK_LEN) == -1)
				{
					LOGSYS_PERROR (errno, instance->totemdisk_log_level_debug, "read error");
					continue;
				}

				if(*(unsigned int*)buf != 1){
				//	write_sector[j]++;
					break;
		//			continue;
				}

		//		if(*(unsigned int*)(buf + sizeof(int)) == instance->totem_config->node_id)
		//			continue;

			//	id =*(unsigned int*)(buf + 4);

				//timestamp = *(long int *)(buf + 2*sizeof(int));
				tv = *(struct timeval *)(buf + 2 * sizeof(int));
				if ((tv.tv_sec < pre_sec[j])||((tv.tv_sec ==  pre_sec[j])&&(tv.tv_usec <= pre_usec[j])))
				{
					read_error_flag += 1;	
					if(read_error_flag >= 5)
						break;	
					else{	
					  	continue;
					}
				}
				read_error_flag = 0;
				//if(timestamp <= pre_time[0])
				//	continue;
				//pre_time[0] = timestamp;

				pre_sec[j] = tv.tv_sec;
				pre_usec[j] = tv.tv_usec;
				if (pre_sec[j] == 0){
					continue;
				}


				msg_len = *(unsigned int*)(buf + 2 * sizeof(int) + sizeof(struct timeval));

			//	log_printf(LOGSYS_LEVEL_CRIT, "&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&:%d,%d,%ld",msg_len,*(buf+4),timestamp);
				if (crypto_encrypt_and_sign (
					instance->crypto_inst,
					(const unsigned char *)(buf + 3 * sizeof(int) + sizeof(struct timeval)),
					msg_len,
					buf_out,
					&buf_out_len) != 0) {
					log_printf(LOGSYS_LEVEL_CRIT, "Error encrypting/signing packet (non-critical)");
					return;
				}

				iovec.iov_base = (void *)buf_out;
				iovec.iov_len = buf_out_len;
				read_max_sector = i;
				res = socket_write(instance, iovec.iov_base, iovec.iov_len);
				if(res < 0)
				{
					log_printf(LOGSYS_LEVEL_CRIT, "write error-%d\n", errno);
				}
			}

//			if (read_max_sector != 0);
			read_sector[j] += read_max_sector;
			
		}
		close(fd);
//		}
	}

	qb_loop_timer_add (instance->totemdisk_poll_handle, QB_LOOP_MED,
		10000*QB_TIME_NS_IN_MSEC,
		(void *)instance,
		timer_function_disk_read_timeout,
		&instance->timer_netif_disk_read);

	return;

}

/*
 * If the interface is up, the sockets for totem are built.  If the interface is down
 * this function is requeued in the timer list to retry building the sockets later.
 */
static void timer_function_netif_check_timeout (
	void *data)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)data;
	int interface_up = 0;
	int flags;
        int res;

	instance->totem_interface->boundto.nodeid = instance->totem_config->node_id;
	instance->totem_interface->boundto.family = 2;
	memcpy(instance->totem_interface->boundto.addr, &instance->totem_config->node_id, sizeof(instance->totem_config->node_id));
	/*
	 * Build sockets for every interface
	 */
	diskif_determine (instance,&interface_up);
	/*
	 * If the network interface isn't back up and we are already
	 * in loopback mode, add timer to check again and return
	 */
	if ((instance->netif_bind_state == BIND_STATE_LOOPBACK &&
		interface_up == 0) ||

	(instance->my_memb_entries == 1 &&
		instance->netif_bind_state == BIND_STATE_REGULAR &&
		interface_up == 1)) {

		qb_loop_timer_add (instance->totemdisk_poll_handle,
			QB_LOOP_MED,
			instance->totem_config->downcheck_timeout*QB_TIME_NS_IN_MSEC,
			(void *)instance,
			timer_function_netif_check_timeout,
			&instance->timer_netif_check_timeout);

		/*
		 * Add a timer to check for a downed regular interface
		 */
		return;
	}

        if (instance->totemdisk_sockets.local_mcast_loop[0] > 0) {
                qb_loop_poll_del (instance->totemdisk_poll_handle,
                        instance->totemdisk_sockets.local_mcast_loop[0]);
                close (instance->totemdisk_sockets.local_mcast_loop[0]);
        }
        if (instance->totemdisk_sockets.local_mcast_loop[1] > 0) {
                close (instance->totemdisk_sockets.local_mcast_loop[1]);
	}

	if (interface_up == 0) {
		/*
		 * Interface is not up
		 */
		instance->netif_bind_state = BIND_STATE_LOOPBACK;
		//bind_address = &localhost;
		totemip_localhost (2, &localhost);
		localhost.nodeid = instance->totem_config->node_id;
		totemip_copy (&instance->my_id, &localhost);
		instance->target_id = instance->my_id.nodeid;

		/*
		 * Add a timer to retry building interfaces and request memb_gather_enter
		 */
		qb_loop_timer_add (instance->totemdisk_poll_handle,
			QB_LOOP_MED,
			instance->totem_config->downcheck_timeout*QB_TIME_NS_IN_MSEC,
			(void *)instance,
			timer_function_netif_check_timeout,
			&instance->timer_netif_check_timeout);
	} else {
		/*
		 * Interface is up
		 */
		instance->netif_bind_state = BIND_STATE_REGULAR;
		//bind_address = &instance->totem_interface->bindnet;
		totemip_copy (&instance->my_id, &instance->totem_interface->boundto);
		instance->target_id = instance->my_id.nodeid;

		res = disk_init(instance);
	      	if(res < 0)
		{
               		LOGSYS_PERROR (errno, instance->totemdisk_log_level_debug,
                       		"disk init error");
			return;
		}
	}

        res = pipe(instance->totemdisk_sockets.local_mcast_loop);
        if (res < 0)
                return;

        flags = fcntl(instance->totemdisk_sockets.local_mcast_loop[0], F_GETFL, 0);
        res = fcntl(instance->totemdisk_sockets.local_mcast_loop[0], F_SETFL, flags | O_NONBLOCK);
        if (res < 0) {
                return;
        }

        flags = fcntl(instance->totemdisk_sockets.local_mcast_loop[1], F_GETFL, 0);
        res = fcntl(instance->totemdisk_sockets.local_mcast_loop[1], F_SETFL, flags | O_NONBLOCK);
        if (res < 0) {
                return;
        }

	qb_loop_poll_add (
		instance->totemdisk_poll_handle,
		QB_LOOP_MED,
		instance->totemdisk_sockets.local_mcast_loop[0],
		POLLIN, instance, disk_deliver_fn);

	qb_loop_timer_add (instance->totemdisk_poll_handle,
		QB_LOOP_MED,
		instance->totem_config->downcheck_timeout*QB_TIME_NS_IN_MSEC,
		(void *)instance, timer_function_disk_read_timeout,
		&instance->timer_netif_disk_read);

	totemip_copy (&instance->my_id, &instance->totem_interface->boundto);
	instance->target_id = instance->my_id.nodeid;

	/*
	 * This reports changes in the interface to the user and totemsrp
	 */
	if (instance->netif_bind_state == BIND_STATE_REGULAR) {
		if (instance->netif_state_report & NETIF_STATE_REPORT_UP) {
			log_printf (instance->totemdisk_log_level_notice,
				"The diskdev interface [%s] is now up.",
				instance->totem_config->diskdev);
			instance->netif_state_report = NETIF_STATE_REPORT_DOWN;
			instance->totemdisk_iface_change_fn (instance->context, &instance->my_id);
		}
		/*
		 * Add a timer to check for interface going down in single membership
		 */
		if (instance->my_memb_entries == 1) {
			qb_loop_timer_add (instance->totemdisk_poll_handle,
				QB_LOOP_MED,
				instance->totem_config->downcheck_timeout*QB_TIME_NS_IN_MSEC,
				(void *)instance,
				timer_function_netif_check_timeout,
				&instance->timer_netif_check_timeout);
		}

	} else {
		if (instance->netif_state_report & NETIF_STATE_REPORT_DOWN) {
			log_printf (instance->totemdisk_log_level_notice,
				"The diskdev interface [%s] is down.", instance->totem_config->diskdev);
			instance->totemdisk_iface_change_fn (instance->context, &instance->my_id);
		}
		instance->netif_state_report = NETIF_STATE_REPORT_UP;

	} 
}

/*
 * Totem Network interface - also does encryption/decryption
 * depends on poll abstraction, POSIX, IPV4
 */

/*
 * Create an instance
 */
int totemdisk_initialize (
	qb_loop_t *poll_handle,
	void **disk_context,
	struct totem_config *totem_config,
	totemsrp_stats_t *stats,
	int interface_no,
	void *context,

	void (*deliver_fn) (
		void *context,
		const void *msg,
		unsigned int msg_len),

	void (*iface_change_fn) (
		void *context,
		const struct totem_ip_address *iface_address),

	void (*target_set_completed) (
		void *context))
{
	struct totemdisk_instance *instance;

	instance = malloc (sizeof (struct totemdisk_instance));
	if (instance == NULL) {
		return (-1);
	}

	totemdisk_instance_initialize (instance);

	instance->totem_config = totem_config;
	instance->stats = stats;

	/*
	* Configure logging
	*/
	instance->totemdisk_log_level_security = 1; //totem_config->totem_logging_configuration.log_level_security;
	instance->totemdisk_log_level_error = totem_config->totem_logging_configuration.log_level_error;
	instance->totemdisk_log_level_warning = totem_config->totem_logging_configuration.log_level_warning;
	instance->totemdisk_log_level_notice = totem_config->totem_logging_configuration.log_level_notice;
	instance->totemdisk_log_level_debug = totem_config->totem_logging_configuration.log_level_debug;
	instance->totemdisk_subsys_id = totem_config->totem_logging_configuration.log_subsys_id;
	instance->totemdisk_log_printf = totem_config->totem_logging_configuration.log_printf;

	/*
	* Initialize random number generator for later use to generate salt
	*/
	instance->crypto_inst = crypto_init (totem_config->private_key,
			totem_config->private_key_len,
			totem_config->crypto_cipher_type,
			totem_config->crypto_hash_type,
			instance->totemdisk_log_printf,
			instance->totemdisk_log_level_security,
			instance->totemdisk_log_level_notice,
			instance->totemdisk_log_level_error,
			instance->totemdisk_subsys_id);
	if (instance->crypto_inst == NULL) {
		free(instance);
		return (-1);
	}
	/*
	 * Initialize local variables for totemdisk
	 */
	instance->totem_interface = &totem_config->interfaces[interface_no];
	totemip_copy (&instance->mcast_address, &instance->totem_interface->mcast_addr);
	memset (instance->iov_buffer, 0, FRAME_SIZE_MAX);

	instance->totemdisk_poll_handle = poll_handle;

	instance->totem_interface->bindnet.nodeid = instance->totem_config->node_id;

	instance->context = context;
	instance->totemdisk_deliver_fn = deliver_fn;

	instance->totemdisk_iface_change_fn = iface_change_fn;

	instance->totemdisk_target_set_completed = target_set_completed;

	localhost.nodeid = instance->totem_config->node_id;

	/*
	 * RRP layer isn't ready to receive message because it hasn't
	 * initialized yet.  Add short timer to check the interfaces.
	 */
	qb_loop_timer_add (instance->totemdisk_poll_handle,
		QB_LOOP_MED,
		100*QB_TIME_NS_IN_MSEC,
		(void *)instance,
		timer_function_netif_check_timeout,
		&instance->timer_netif_check_timeout);

	*disk_context = instance;
	return (0);
}

void *totemdisk_buffer_alloc (void)
{
	return malloc (FRAME_SIZE_MAX);
}

void totemdisk_buffer_release (void *ptr)
{
	return free (ptr);
}

int totemdisk_processor_count_set (
	void *disk_context,
	int processor_count)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	int res = 0;

	instance->my_memb_entries = processor_count;
	qb_loop_timer_del (instance->totemdisk_poll_handle,
		instance->timer_netif_check_timeout);
	if (processor_count == 1) {
		qb_loop_timer_add (instance->totemdisk_poll_handle,
			QB_LOOP_MED,
			instance->totem_config->downcheck_timeout*QB_TIME_NS_IN_MSEC,
			(void *)instance,
			timer_function_netif_check_timeout,
			&instance->timer_netif_check_timeout);
	}

	return (res);
}

int totemdisk_recv_flush (void *disk_context)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	struct pollfd ufd;
	int nfds;
	int sock;
	int res = 0;

	instance->flushing = 1;

		sock = -1;
		sock = instance->totemdisk_sockets.local_mcast_loop[0];
		assert(sock != -1);

		do {
			ufd.fd = sock;
			ufd.events = POLLIN;
			nfds = poll (&ufd, 1, 0);
			if (nfds == 1 && ufd.revents & POLLIN) {
				disk_deliver_fn (sock, ufd.revents, instance);
			}
		} while (nfds == 1);

	instance->flushing = 0;

	return (res);
}

int totemdisk_send_flush (void *disk_context)
{
	return 0;
}

int totemdisk_token_send (
	void *disk_context,
	const void *msg,
	unsigned int msg_len)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	int res = 0;

//	if (instance->token_target.nodeid == instance->target_id) {
		cast_sendmsg (instance, msg, msg_len);
  //      }else{
		write_disk (instance,msg,msg_len);
//	}

	return res;
}

int totemdisk_mcast_flush_send (
	void *disk_context,
	const void *msg,
	unsigned int msg_len)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	int res = 0;

	cast_sendmsg (instance, msg, msg_len);
	write_disk(instance,msg,msg_len);

	return res;
}

int totemdisk_mcast_noflush_send (
	void *disk_context,
	const void *msg,
	unsigned int msg_len)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	int res = 0;

	cast_sendmsg (instance, msg, msg_len);
	write_disk(instance,msg,msg_len);

	return res;
}

extern int totemdisk_iface_check (void *disk_context)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	int res = 0;

	timer_function_netif_check_timeout (instance);

	return (res);
}

extern void totemdisk_net_mtu_adjust (void *disk_context, struct totem_config *totem_config)
{
	return;

	assert(totem_config->interface_count > 0);

	totem_config->net_mtu -= crypto_sec_header_size(totem_config->crypto_cipher_type,
							totem_config->crypto_hash_type) +
				 totemip_udpip_header_size(totem_config->interfaces[0].bindnet.family);
}

const char *totemdisk_iface_print (void *disk_context)  {
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	const char *ret_char;

	ret_char = instance->totem_config->diskdev;
	return (ret_char);
}

int totemdisk_iface_get (
	void *disk_context,
	struct totem_ip_address *addr)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	int res = 0;

	memcpy (addr, &instance->my_id, sizeof (struct totem_ip_address));

	return (res);
}

int totemdisk_token_target_set (
	void *disk_context,
	const struct totem_ip_address *token_target)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	int res = 0;

	memcpy (&instance->token_target, token_target,
		sizeof (struct totem_ip_address));

	instance->totemdisk_target_set_completed (instance->context);

	return (res);
}

extern int totemdisk_recv_mcast_empty (
	void *disk_context)
{
	struct totemdisk_instance *instance = (struct totemdisk_instance *)disk_context;
	unsigned int res;
	struct pollfd ufd;
	int nfds;
	int msg_processed = 0;
	int sock;
	struct iovec *iovec;
	
	iovec = &instance->totemdisk_iov_recv_flush;

		sock = -1;
		sock = instance->totemdisk_sockets.local_mcast_loop[0];
		assert(sock != -1);
		do {
			ufd.fd = sock;
			ufd.events = POLLIN;
			nfds = poll (&ufd, 1, 0);

			if (nfds == 1 && ufd.revents & POLLIN) {
				res = disk_recvmsg (sock, iovec->iov_base, iovec->iov_len);
				if (res > 0) {
					msg_processed = 1;
				} else {
					msg_processed = -1;
				}
			}
		} while (nfds == 1);

        return (msg_processed);
}
