/*
 * Copyright (c) 2005 MontaVista Software, Inc.
 * Copyright (c) 2006-2011 Red Hat, Inc.
 *
 * All rights reserved.
 *
 * Author: Steven Dake (sdake@redhat.com)
 *
 * This software licensed under BSD license, the text of which follows:
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the MontaVista Software, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef TOTEMDISK_H_DEFINED
#define TOTEMDISK_H_DEFINED

#include <sys/types.h>
#include <sys/socket.h>
#include <qb/qbloop.h>

#include <corosync/totem/totem.h>
#include <stdio.h>
/**
 * Create an instance
 */
extern int totemdisk_initialize (
	qb_loop_t* poll_handle,
	void **disk_context,
	struct totem_config *totem_config,
	totemsrp_stats_t *stats,
	int interface_no,
	void *context,

	void (*deliver_fn) (
		void *context,
		const void *msg,
		unsigned int msg_len),

	void (*iface_change_fn) (
		void *context,
		const struct totem_ip_address *iface_address),

	void (*target_set_completed) (
		void *context));

extern void *totemdisk_buffer_alloc (void);

extern void totemdisk_buffer_release (void *ptr);

extern int totemdisk_processor_count_set (
	void *disk_context,
	int processor_count);

extern int totemdisk_token_send (
	void *disk_context,
	const void *msg,
	unsigned int msg_len);

extern int totemdisk_mcast_flush_send (
	void *disk_context,
	const void *msg,
	unsigned int msg_len);

extern int totemdisk_mcast_noflush_send (
	void *disk_context,
	const void *msg,
	unsigned int msg_len);

extern int totemdisk_recv_flush (void *disk_context);

extern int totemdisk_send_flush (void *disk_context);

extern int totemdisk_iface_check (void *disk_context);

extern int totemdisk_finalize (void *disk_context);

extern void totemdisk_net_mtu_adjust (void *disk_context, struct totem_config *totem_config);

extern const char *totemdisk_iface_print (void *disk_context);

extern int totemdisk_iface_get (
	void *disk_context,
	struct totem_ip_address *addr);

extern int totemdisk_token_target_set (
	void *disk_context,
	const struct totem_ip_address *token_target);

extern int totemdisk_crypto_set (
	void *disk_context,
	const char *cipher_type,
	const char *hash_type);

extern int totemdisk_recv_mcast_empty (
	void *disk_context);

#endif /* TOTEMDISK_H_DEFINED */
